/*
 * cmfgetln.c
 *  
 *
 * Aternatives:
 * 1.  fgets
 *		Pro:  Ubiquitous
 *		Con:  Buffer size issues
 * 2.  C++ iostream getline
 *		Pro:  Buffer OK using C++ string
 *		Con:  Slow, requires iostream and C++
 * 3.  fgetln/getline
 *		Pro:  OK buffer, speed swell, uses FILE *, char * types
 *		Con:  BSD fgetln not on GNU systems, vice versa
 *
 */
#include <sys/param.h> 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmfgetln.h"


static unsigned short int Bfs = 40;

/*
 * cmfgetln is the easy-to-use version of cmfgettoch for 
 * getting a new line quick.  See notes before cmfgettoch.
 */
char *cmfgetln(FILE *ifp)
{

    return cmfgettoch(ifp, '\n');
}


/*
 * cmfgettoch will read from the FILE pointer ifp until the character to is found, and will 
 * return a string of all the data thereto, replacing the character to with a \0 character.  
 * Will return NULL if called during an EOF condition.
 * 
 * Buffers allocated here should be freed by whatever function uses it.
 */
char *cmfgettoch(FILE *ifp, const char to)
{
    #ifdef BSD	/* create BSD optimized cmfgettoch */
    register unsigned char *pos;
    char *ret;
    extern int __srefill __P((FILE *));
    register size_t length;
    

    /*
     * Do we have any more input?  If we haven't,
     * we should probably return some sort of value 
     * stating that.
     */
    if (ifp->_r <= 0 && __srefill(ifp))
        return NULL;
                
    /*
     * So, we search for our character.  If one is found, 
     * we figure the length of the string in relation
     * to fp->_p.  If not, we figure it in relation to 
     * fp->_r (what's left in the file).
     * If you're confused, /usr/include/stdio.h has the 
     * FILE type definition on Berkeley systems.
     */
    if ((pos = memchr((void *)ifp->_p, to, ifp->_r )))
        length = (++pos - ifp->_p);
    else
        length = (ifp->_r + 1);

    if((ret = (char *)malloc(length * sizeof(char))) == NULL) {
        perror(NULL);	/* bail out, no more VM */
        exit(1);
    }
    ret = memcpy(ret, (void *)ifp->_p, length);
    ret[length - 1] = '\0';
    ifp->_p = pos;
    ifp->_r -= length;
    return ret;
    
    #else	/* we're not on BSD, default to fgetc */
    register int ch;
    register char *buf;
    register size_t blen, cnt;
    
    
    cnt = 0;
    blen = Bfs;
    buf = (char *)malloc(blen);
    
    while (((ch = fgetc(ifp)) != to) || (ch == EOF)) {
        buf[cnt] = ch;
        
        /* whoa...end of file.  Time to stop */
        if (ch == EOF)
            break;
            
        cnt++;
        
        /* out of buffer space...reallocate */
        if (cnt >= blen) {
            blen += Bfs;
            if ((buf = (char *)realloc(buf, blen)) == NULL) {
                    perror(NULL);
                    exit(1);
            }
        }
    }
    
    /* Return NULL terminated string */
    buf[cnt] = '\0';
    
    /* No more lines...return NULL pointer */
    if((cnt == 0) && (ch == EOF)) {
        free(buf);
        return NULL;
    }
    
    /* All is OK, return the string */
    return buf;
    #endif
}

/*
 * cmtunebfs will set how large of a buffer to reallocate every time a new 
 * buffer is created or needs to be reallocated.  By varying this number, 
 * one can carefully adjust ones needs of processor usage via realloc versus 
 * memory wastage by malloc-ing too much RAM.
 *
 * Although cmtunebfs may reject the value passed to it, it won't return 
 * an error condition, it will continue working, but with the previous value 
 * therein.
 *
 * We do reject all values less than 16, most of these really hurt performance.  
 * It's suggested that immensely huge values are not used, so we cut those off at 
 * 256.  If you find a circumstance in which these values are daft, well, feel 
 * free to edit this.
 */
void cmtunebfs(const unsigned short int newbfs)
{

    if (newbfs < 16)
        return;
    
    if (newbfs > 256) {
        Bfs = 256;
        return;
    }
    
    Bfs = newbfs;
}